#include "CAalgos.h"



/* Simple channel assignment where all the nodes
   have one radio assigned the same channel
*/
double OneChannel::Run (Topology *T)
{
//	Topology* T = new Topology(*Gi);     /* copy of the initial graph */
	T->ResetChannels();

	//for each node, one of the radios is assigned channel 1
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		u->info->radio[0].channel = 1; 

	// each node can reach its neighbors through that radio
	// (this implies that the channel of each link is set to 1)
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			v->radio = &u->info->radio[0];

	// compute Utot for each link (initially the collision domain equals the neighborhood)
	// for every directed link u->v
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			v->Utot = T->computeUtot(TLinkItrs(u,v));

	return 0;
}



CAalgo* CAalgo::Create (string algo, const list<pair<string,string> >& params)
{
  if (algo == "DoNothing")
    return new CAalgo;
  if (algo == "OneChannel")
    return new OneChannel;
  if (algo == "IATC")
    return new IATC;
  if (algo == "LACA")
    return new LACA;
  if (algo == "FCPRA")
    return new FCPRA (params);
  if (algo == "MVCRA")
    return new MVCRA (params);
  if (algo == "MPCARA")
    return new MPCARA (params);
  if (algo == "MPCARA_OPT")
    return new MPCARA_OPT (params);
  if (algo == "PACA2")
    return new PACA2 (params);
  if (algo == "PACA")
    return new PACA (params);
  if (algo == "E2CAR_OPT")
    return new E2CAR_OPT (params);

  return NULL;
}

