#include <algorithm>
#include <cassert>
#include <ext/functional>
#include "WMtypes.h"
#include "CAalgos.h"
#include "pfr_methods.h"
#include <iostream>

using namespace std;


/* estimates the capacity of each link and store it both in availBW and flow */
double EstimateCapacity(Topology *T, TLinkItrs l)
{
	double sumflows = 0.;

	vector<TLinkItrs> Dcoll = T->computeNeighborhood(l);

	for (vector<TLinkItrs>::iterator i=Dcoll.begin(); i!=Dcoll.end(); i++)
		if (i->second->radio->channel == l.second->radio->channel)
			sumflows += i->second->flow;

	if (sumflows)
		return l.second->flow / sumflows * l.second->rate;
	else
		return 0.;
}



void RanGopChi(Topology *T, int NumFlows, RndVar *Source, RndVar *Dest, RndVar *Bwdr, double seed)
{
	double CSgoodput;
	int curiter = 0, maxiter = 10;

	PFR_method* pfr = new PFR_Raniwala(Bwdr, seed);
	pfr->pfr_calc(T);

	do
	{
		LACA laca;
		laca.Run(T);

		// estimate capacity for every link
		for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
			for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
				v->availBW = EstimateCapacity(T,TLinkItrs(u,v));
	
		// store the estimated capacity in flow
		for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
			for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
				v->flow = 2 * v->availBW;
	
		CSgoodput = T->RouteFlows(NumFlows, Source, Dest, Bwdr, seed);
	
		// compute the allocated bandwidth, which serves as the new flow
		for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
			for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
				v->flow -= v->availBW;

		curiter++;

	} while (T->getUtotMax().second > 1. && curiter < maxiter);

	T->scaleFlowsByUtotMax();
}


