#include "WMtypes.h"

#include <math.h>
#include <iostream>
#include <set>
#include <map>
#include <algorithm>
#include <numeric>
#include <functional>
#include <cassert>

#define CHANNEL(l)              \
l.second->radio->channel



/* Compute the new Utot for every link involved (except l itself) in case c is assigned to l */
void fcra_computeNewUtot(Topology *T, TLinkItrs l, unsigned int c);

/* Binary predicate needed to find the link with the maximum Ftot */
//inline bool max_flow(const pair<int,int>& x, const pair<int,int>& y) { return x.second < y.second; }

/* Binary function needed to comput the sum of the rates for all links u->v */
// inline float sumRate(float result, TGraphNeighbor i) { return result + i.rate; }

/* This function is invoked when we extract i<->j and want to optimize the radios of i (i.e., remove channel 1)*/
/* Returns true if channel 1 is removed */
bool fcra_OptimizeRadios(Topology *T, TLinkItrs l, int i, set<TLinkItrs> Q, bool turbo);


/* Given the initial graph, performs a channel assignment
   and returns the induced graph
*/
void FCRA(Topology *T, bool Optimize, bool turbo, bool RateAdapt)
{
	/* INITIALIZATION */
	//for each node, one of the radios is assigned channel 1
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		u->info->radio[0].channel = 1; 

	// each node can reach its neighbors through that radio
	// (this implies that the channel of each link is set to 1)
	// also set the rate of every link to the highest rate
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
		{
			v->radio = &u->info->radio[0];
			if (RateAdapt)
				v->rate = availTXrates[NUMRATES-1];
		}
//T->node[1].neighbor[0].rate = 12.;
//T->node[3].neighbor[0].rate = 24.;
//solveMaxFtoU(T, T->throughput);
	// compute Utot for each link (initially the collision domain equals the neighborhood)
	// for every directed link u->v
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			v->Utot = T->computeUtot(TLinkItrs(u,v));

	// Set containing a pointer to all the directed links
	set<TLinkItrs> Q;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
				Q.insert(TLinkItrs(u,v));

	/* END INITIALIZATION */

	set<TLinkItrs>::iterator ptr_l;
	TLinkItrs l;
	double max_Utot;
	bool ch1removed = false;
	unsigned int l_channel;
	TLinkItrs invLink;

	while (!Q.empty())
	{
		// Extract the link with the maximum Utot from Q
		max_Utot = -1.;
		for (set<TLinkItrs>::iterator p=Q.begin(); p!=Q.end(); p++)
			if (p->second->Utot > max_Utot)
			{
				max_Utot = p->second->Utot;
				ptr_l = p;
			}

		// Erase the extracted link from Q
		l = *ptr_l;
		Q.erase(ptr_l);
		l_channel = CHANNEL(l);
		if (Optimize)
			ch1removed = fcra_OptimizeRadios(T, l, l.first->info->id, Q, turbo) || fcra_OptimizeRadios(T, l, l.second->info->id, Q, turbo);

		/* if the end nodes only share channel 1 and the optimization step removes channel 1 from one of the nodes
		   remove the inverse link (by assigning channel 0) and updates the Utot of the links involved */
		if (ch1removed)
		{
			invLink = T->FindInverseLink(l);
			RadioIface *fakeradio = new RadioIface();
			invLink.second->radio = fakeradio;
			invLink.second->radio->channel = (unsigned int)1;
			fcra_computeNewUtot(T, invLink, 0);
			invLink.second->radio->channel = 0;
		}

		set<unsigned int> uChannels, vChannels, S, Sc;
		S.clear();
		Sc.clear();
		// Find the set of channels of the end nodes of the link to be extracted
// 		findChannels(l.first, uChannels);
// 		findChannels(l.second, vChannels);
		uChannels = l.first->info->findChannels();
		vChannels = l.second->info->findChannels();

		// Consider all the cases
		// if both end nodes have available radios
		if ( uChannels.size()<l.first->info->numAvailRadios() && vChannels.size()<l.second->info->numAvailRadios() )
			// put all the channels in S
			for (unsigned int i=1; i<=T->numAvailChannels; S.insert(i++));
		// if u has all radios assigned a channel
		else if ( uChannels.size()==l.first->info->numAvailRadios() && vChannels.size()<l.second->info->numAvailRadios() )
			// put the channels of u in S
			copy(uChannels.begin(), uChannels.end(), inserter(S, S.begin()));
		// if v has all radios assigned a channel
		else if ( uChannels.size()<l.first->info->numAvailRadios() && vChannels.size()==l.second->info->numAvailRadios() )
			// put the channels of v in S
			copy(vChannels.begin(), vChannels.end(), inserter(S, S.begin()));
		// if all the radios of both end nodes have been assigned a channel
		else
			// put the common channels of u and v in S
			set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(S, S.begin()));

		// S is now the set of candidate channels to be assigned to the link to be extracted
		assert( !S.empty() );

		// put the common channels of u and v in Sc
		set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(Sc, Sc.begin()));

		map<unsigned int, double> Umax;

		// for each channel c in S
		for (set<unsigned int>::iterator c=S.begin(); c!=S.end(); c++)
			Umax[*c] = 0.;

		// compute Umax[c] for all c in S
		// for every directed link x->y
		for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
			for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
				// whose potential collision domain includes the extracted link and having a channel in S
				if ((TLinkItrs(x,y) != l) && (T->PotInterfLinks(l, TLinkItrs(x,y))) && (S.find(y->radio->channel)!=S.end()))
				{
					/* remove the contribution of the extracted link from the Utot of the links having the same channel */
					if (y->radio->channel == l_channel)
						y->Utot -= l.second->flow/l.second->rate;
					if (y->Utot > Umax[y->radio->channel])
						Umax[y->radio->channel] = y->Utot;
				}

		map<unsigned int, float> rate;
		unsigned int txpower = l.second->txpower;
		int m, mmin;
		double min,save;

		/* we use a fake radio to compute Utot(u->v) for every channel */
		RadioIface *fakeradio = new RadioIface();
		l.second->radio = fakeradio;

		// for each channel c in S
		for (set<unsigned int>::iterator c=S.begin(); c!=S.end(); c++)
		{
			// we overwrite the channel and the rate of the extracted link to compute its Utot correctly
			CHANNEL(l) = *c;
			if (RateAdapt)
			{
				l.second->rate = availTXrates[NUMRATES-1];
	
				mmin = (m = NUMRATES-1);
				min = T->computeUtot(l) - l.second->flow/l.second->rate;
	
				while ((m>0) && (min>Umax[*c]))
				{
					l.second->rate = availTXrates[--m];
					if ( (save = T->computeUtot(l) - l.second->flow/l.second->rate) < min)
					{
						min = save;
						mmin = m;
					}
				}
				// this rate will be used in case a link u-> v is established on this channel
				rate[*c] = availTXrates[mmin];
			}
			else
			{
				min = T->computeUtot(l) - l.second->flow/l.second->rate;
				rate[*c] = l.second->rate;
			}

			if (min > Umax[*c])
				Umax[*c] = min;
		}

		delete fakeradio;

		set<unsigned int> C;
		set_difference(S.begin(), S.end(), Sc.begin(), Sc.end(), inserter(C, C.begin()));

		/* leave in Umax only the channels in Sc and the channel in S-Sc that minimizes Umax */
		if (!C.empty())
		{
			set<unsigned int>::iterator c = C.begin();
			min = Umax[*c];
			for (set<unsigned int>::iterator c0=++C.begin(); c0!=C.end(); c0++)
				if (Umax[*c0]<min)
				{
					Umax.erase(*c);
					c = c0;
					min = Umax[*c0];
				} 
				else
					Umax.erase(*c0);
		}

		double F = l.second->flow;
		TGraphNeighbor *newlink;
		list<TGraphNeighbor>::iterator currlink;

		MeshNode* backup_info = l.second->info;
		/* remove v from among u's neighbors, i.e., remove the extracted link */
		l.first->neighbor.erase(l.second);

		unsigned int c;
		float sumrates;
		double base, I;
		bool done = false;


		do
		{
			/* add a new link u-> v on the channel that minimizes Umax */
			newlink = new TGraphNeighbor(backup_info);

			l.first->neighbor.push_back(*newlink);

			currlink = --l.first->neighbor.end();

			c = min_element(Umax.begin(), Umax.end(), compareSecond<unsigned int, double>())->first;

			T->AssignChannel2Link( TLinkItrs(l.first,currlink), c );
			currlink->rate = rate[c];               /* set the previously computed rate */
			currlink->txpower = txpower;                /* set the previously computed power */

			sumrates = 0.;
			for (list<TGraphNeighbor>::iterator v=l.first->neighbor.begin(); v!=l.first->neighbor.end(); v++)
				if (v->info->id == currlink->info->id)
					sumrates+=v->rate;

			base = Umax[c];
			Umax.erase(c);

			if ( (!Umax.empty()) && (F>=(min_element(Umax.begin(),Umax.end(),compareSecond<unsigned int, double>())->second - base)*sumrates) )
				I = min_element(Umax.begin(),Umax.end(),compareSecond<unsigned int, double>())->second - base;
			else
			{
				I = F / sumrates;
				done = true;
			}

			for (list<TGraphNeighbor>::iterator k=l.first->neighbor.begin(); k!=l.first->neighbor.end(); k++)
				if (k->info->id == currlink->info->id)
				{
					k->flow += I * k->rate;
					F -= I * k->rate;
				}

		} while (!done);

		if (ch1removed)
		// assign the same channel as the extracted link to the inverse link
		{
			fcra_computeNewUtot(T, invLink, c);
			T->AssignChannel2Link( invLink, c );
			invLink.second->Utot = T->computeUtot(invLink);
		}

		// compute the Utot for the newly established links
		for (list<TGraphNeighbor>::iterator k=l.first->neighbor.begin(); k!=l.first->neighbor.end(); k++)
			if (k->info->id == currlink->info->id)
				k->Utot = T->computeUtot(TLinkItrs(l.first,k));

		// Update the Utot for all the other links involved
		// for every directed link x->y
		for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
			for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
				// for every newly establisthed link u->v
				for (list<TGraphNeighbor>::iterator k=l.first->neighbor.begin(); k!=l.first->neighbor.end(); k++)
					if ( (k->info->id == currlink->info->id) && ( !ch1removed || TLinkItrs(x,y) != invLink ) &&
					// such that the potential collision domain of x->y includes the new link and have the same channel
					     (x!=l.first || y!=k) && (T->PotInterfLinks(TLinkItrs(l.first,k), TLinkItrs(x,y))) && (y->radio->channel==k->radio->channel) )
						/* add the contribution of the new link to the Utot of x->y */
						y->Utot += k->flow/k->rate;
	}

}



/* Compute the new Utot for every link involved (except l itself) in case c is assigned to l */
void fcra_computeNewUtot(Topology *T, TLinkItrs l, unsigned int c)
{
	/* the new Utot for the other links is obtained from the current Utot value */
	// for every directed link x->y
	for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
		for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
			// such that the potential collision domain of x->y includes l
			if ( (l!=TLinkItrs(x,y)) && (T->PotInterfLinks(l,TLinkItrs(x,y))) )
			{
				/* by subtracting the F/C ratio of l if l escapes the collision domain of l0 */
				if ((y->radio->channel == CHANNEL(l)) && (y->radio->channel != c))
					y->Utot -= l.second->flow / l.second->rate;
				/* by adding the F/C ratio of l if l joins the collision domain of l0 */
				else if ((y->radio->channel != CHANNEL(l)) && (y->radio->channel == c))
					y->Utot += l.second->flow / l.second->rate;
			}
}


/* This function is invoked when we extract link l and want to optimize the radios of i (i.e., remove channel 1)
   */
bool fcra_OptimizeRadios(Topology *T, TLinkItrs l, int i, set<TLinkItrs> Q, bool turbo)
{
	bool invremoved = false;
	set<TLinkItrs> L, extracted, nonextracted;
	set<pair<TLinkItrs,unsigned int> > tomodify;

	set<unsigned int> iChannels, lChannels, S;
// 	findChannels((T->node.begin()+i),iChannels);
	iChannels = T->node[i].info->findChannels();

	/* if i has available radios or no radio is assigned to 1, do nothing */
	if ( (iChannels.size()<T->node[i].info->numAvailRadios()) || (find(iChannels.begin(),iChannels.end(),(unsigned int)1)==iChannels.end()) )
		return false;

	/* insert all the links leaving or entering i into L */
	for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
		for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
			if ( x->info->id == i || y->info->id == i )
				L.insert(TLinkItrs(x,y));

	L.erase(l);

	/* compute L-Q, i.e., the links that have already been extracted */
	set_difference(L.begin(), L.end(), Q.begin(), Q.end(), inserter(extracted, extracted.begin()));

	/* if an extracted link has been assigned channel 1, we can do nothing */
	for (set<TLinkItrs>::iterator l0=extracted.begin(); l0!=extracted.end(); l0++)
		if (CHANNEL((*l0)) == 1)
			return false;

	/* compute L\cap Q, i.e., the links that have not been extracted yet */
	set_difference(L.begin(), L.end(), extracted.begin(), extracted.end(), inserter(nonextracted, nonextracted.begin()));


	list<TGraphNeighbor>::iterator invL;
	if (turbo && (invL = T->find_neighbor(l.second->info->id, l.first->info->id)) != T->node[l.second->info->id].neighbor.end() )
	{
		bool inverse = ( find(nonextracted.begin(), nonextracted.end(), TLinkItrs(T->node.begin()+l.second->info->id, invL)) != nonextracted.end() );
		if (inverse)
		{
// 			findChannels((T->node.begin()+l.first->info->id),iChannels);
// 			findChannels((T->node.begin()+l.second->info->id),lChannels);
			iChannels = T->node[l.first->info->id].info->findChannels();
			lChannels = T->node[l.second->info->id].info->findChannels();
			S.clear();
			set_intersection(iChannels.begin(), iChannels.end(), lChannels.begin(), lChannels.end(), inserter(S, S.begin()));
			S.erase(1);
			if (S.empty())
			{
				nonextracted.erase(TLinkItrs(T->node.begin()+l.second->info->id, invL));
				invremoved = true;
			}
		}
	}

	/* if a nonextracted link has no channel but 1 in common with i, we can do nothing */
	for (set<TLinkItrs>::iterator l0=nonextracted.begin(); l0!=nonextracted.end(); l0++)
	{
		S.clear();
		// Find the set of channels of the end nodes of the link to be extracted
// 		findChannels((l0->first), iChannels);
// 		findChannels((l0->second), lChannels);
		iChannels = l0->first->info->findChannels();
		lChannels = l0->second->info->findChannels();

		set_intersection(iChannels.begin(), iChannels.end(), lChannels.begin(), lChannels.end(), inserter(S, S.begin()));
		S.erase(1);
		if (!S.empty())
			tomodify.insert(pair<TLinkItrs,unsigned int>(*l0,*S.begin()));
		else
			return false;
	}

	// if we reach this point, it means that we can remove channel 1 from the radios of i
	/* for each link to be modified */
	for (set<pair<TLinkItrs,unsigned int> >::iterator l0=tomodify.begin(); l0!=tomodify.end(); l0++)
	{
		if (l0->first.second->radio->channel != l0->second)
		{
			/* update Utot for neighbors*/
			fcra_computeNewUtot(T, l0->first, l0->second);
			/* assign the new channel to l0 */
			T->AssignChannel2Link(l0->first,l0->second);
			/* compute Utot for the modified link*/
			l0->first.second->Utot = T->computeUtot(l0->first);
		}
	}

	/* set to channel 0 the radio of i that is set to channel 1, i.e.,make it available */
// 	pair<bool, unsigned int> foundCh = SearchChannel(T->node[i].info, 1);
	RadioIfaceIterator ri = T->node[i].info->findRadioWithChannel (1);

// 	assert( foundCh.first );
	assert ( ri != T->node[i].info->radio.end() );

// 	T->node[i].info->radio[foundCh.second].channel = 0;
	ri->channel = 0;

	return invremoved;
}



