#include "CAalgos.h"

#include <algorithm>
#include <cassert>

typedef pair<list<TGraphNeighbor>::iterator, list<TGraphNeighbor>::iterator> TUndLink;

bool operator<(const TUndLink& x, const TUndLink& y) { return true; }

struct PEcard
{
	bool operator()(const set<TUndLink> &x, const set<TUndLink> &y)
	{
		return x.size() > y.size();
	}
};



void replace(Topology *T, list<TGraphNeighbor>::iterator u, list<TGraphNeighbor>::iterator v, unsigned int k, unsigned int k1)
{
	/* replace k1 with k */
// 	RadioIface *i;
// 	for (i=u->info->radio; i->channel!=k1; i++);
// 	i->channel = k;
	for (vector<RadioIface>::iterator i=u->info->radio.begin(); i!=u->info->radio.end(); i++)
	  if (i->channel == k1)
	  {
	    i->channel = k;
	    break;
	  }

	/* assign u->radio and v->radio */
	for (u->radio=&v->info->radio[0]; u->radio->channel != k; u->radio++);
	for (v->radio=&u->info->radio[0]; v->radio->channel != k; v->radio++);

	for (list<TGraphNeighbor>::iterator w=T->node[u->info->id].neighbor.begin(); w!=T->node[u->info->id].neighbor.end(); w++)
		if ((w->radio!=NULL) && (w!=v) && (w->radio->channel == k))		/* link already considered */
		{
			set<unsigned int> uChannels, wChannels, CommonChannels;

			for (unsigned int r=0; r<u->info->numAvailRadios(); uChannels.insert(u->info->radio[r++].channel));
			for (unsigned int r=0; r<w->info->numAvailRadios(); wChannels.insert(w->info->radio[r++].channel));

			/* compute the intersection between uChannels and wChannels */
			set_intersection(uChannels.begin(), uChannels.end(), wChannels.begin(), wChannels.end(), inserter(CommonChannels, CommonChannels.begin()));
			CommonChannels.erase(0);

			list<TGraphNeighbor>::iterator z;
			for (z=T->node[w->info->id].neighbor.begin(); z->info->id!=u->info->id; z++);

			if (!CommonChannels.empty())
			{
				for (w->radio=&u->info->radio[0]; w->radio->channel != *(CommonChannels.begin()); w->radio++);
				for (z->radio=&w->info->radio[0]; z->radio->channel != *(CommonChannels.begin()); z->radio++);
			}
			else
				replace(T,w,z,k,k1);
		}
}


/* Given the initial graph, performs a channel assignment
   and returns the induced graph
*/
double IATC::Run (Topology *T)
{
//	Topology* T = new Topology(*Gi);     /* copy of the initial graph */

	/* for each link find the set of potentially interfering links */
	multimap<set<TUndLink>, TUndLink, PEcard> PE;		/* the map is sorted in descending order of |PE[e]| */

	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			if (u->info->id < v->info->id)		/* don't count a link twice */
			{
				list<TGraphNeighbor>::iterator node2;
				for (node2=T->node[v->info->id].neighbor.begin(); node2!=T->node[v->info->id].neighbor.end() && node2->info->id!=u->info->id; node2++);
				assert( node2 != T->node[v->info->id].neighbor.end() );

				TUndLink e(v,node2);
				set<TUndLink> PEe;
/* if i<->j potentially interfere with e, PEe only contains the pair <i,j> such that i->info->id > j->info->id */

				/* add all the links outgoing the nodes in the interference range of any of the end nodes of e */
				for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
					for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
						if ( T->PotInterfLinks(TLinkItrs(u,v),TLinkItrs(x,y)) )
						{
							for (node2=T->node[y->info->id].neighbor.begin(); node2!=T->node[y->info->id].neighbor.end() && node2->info->id!=x->info->id; node2++);
							assert( node2!=T->node[y->info->id].neighbor.end() );

							if ( y->info->id > node2->info->id )
								PEe.insert(TUndLink(y,node2));
							else
								PEe.insert(TUndLink(node2,y));
						}

				PE.insert(pair<set<TUndLink>, TUndLink>(PEe, e));
			}

	for (multimap<set<TUndLink>, TUndLink, PEcard>::const_iterator e=PE.begin(); e!=PE.end(); ++e)
	{
		set<unsigned int> uChannels, vChannels, CommonChannels;
		list<TGraphNeighbor>::iterator u=e->second.first, v=e->second.second;	/* end nodes of e */

		for (unsigned int r=0; r<u->info->numAvailRadios(); uChannels.insert(u->info->radio[r++].channel));
		uChannels.erase(0);
		for (unsigned int r=0; r<v->info->numAvailRadios(); vChannels.insert(v->info->radio[r++].channel));
		vChannels.erase(0);
		CommonChannels.clear();

		/* compute the intersection between uChannels and vChannels */
		set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(CommonChannels, CommonChannels.begin()));

		if (!CommonChannels.empty())
		{	/* set u->radio and v->radio */
			for (u->radio=&v->info->radio[0]; u->radio->channel != *(CommonChannels.begin()); u->radio++);
			for (v->radio=&u->info->radio[0]; v->radio->channel != *(CommonChannels.begin()); v->radio++);
		}
		else if ( (CommonChannels.empty()) && (uChannels.size()<u->info->numAvailRadios()) && (vChannels.size()<v->info->numAvailRadios()) )
		{
			vector<int> occurrences(T->numAvailChannels+1);
			for (vector<int>::iterator i=occurrences.begin(); i!=occurrences.end(); i++) *i=0;

			/* find the number of occurrences of each channel in PE[e] */
			for (set<TUndLink>::iterator l=e->first.begin(); l!=e->first.end(); l++)
				if (l->first->radio != NULL)
					occurrences[l->first->radio->channel]++;

			/* find the least used channel in PE[e] */
			unsigned int k = min_element(++occurrences.begin(), occurrences.end()) - occurrences.begin();
			assert(k>0);

			/* assign k to the first unassigned interface on both end nodes */
			if (uChannels.find(k) == uChannels.end())
				u->info->radio[uChannels.size()].channel = k;
			if (vChannels.find(k) == vChannels.end())
				v->info->radio[vChannels.size()].channel = k;

			/* set u->radio and v->radio */
			for (u->radio=&v->info->radio[0]; u->radio->channel != k; u->radio++);
			for (v->radio=&u->info->radio[0]; v->radio->channel != k; v->radio++);
		}
		else if ((CommonChannels.empty()) && (uChannels.size()<u->info->numAvailRadios()) && (vChannels.size()==v->info->numAvailRadios()) )
		{
			vector<int> occurrences(vChannels.size());
			for (vector<int>::iterator i=occurrences.begin(); i!=occurrences.end(); i++) *i=0;

			/* find the number of occurrences of the channels of v in PE[e] */
			set<unsigned int>::iterator pos;
			int j;
			for (set<TUndLink>::iterator l=e->first.begin(); l!=e->first.end(); l++)
				if ( (l->first->radio != NULL) && ((pos=vChannels.find(l->first->radio->channel)) != vChannels.end()) )
				{
					j=0;
					for (set<unsigned int>::iterator temp=vChannels.begin(); temp!=pos; temp++)
						j++;
					occurrences[j]++;
				}

			/* find the least used channel in PE[e] */
			j = min_element(occurrences.begin(), occurrences.end()) - occurrences.begin();
			pos = vChannels.begin();
			for (int i=0; i<j; i++) pos++;
			unsigned int k = *pos;
			assert(k>0);

			/* assign k to the first unassigned interface on u->info->id */
			u->info->radio[uChannels.size()].channel = k;

			/* set u->radio and v->radio */
			for (u->radio=&v->info->radio[0]; u->radio->channel != k; u->radio++);
			v->radio = &(u->info->radio[uChannels.size()]);
		}
		else if ((CommonChannels.empty()) && (uChannels.size()==u->info->numAvailRadios()) && (vChannels.size()<v->info->numAvailRadios()) )
		{
			vector<int> occurrences(uChannels.size());
			for (vector<int>::iterator i=occurrences.begin(); i!=occurrences.end(); i++) *i=0;

			/* find the number of occurrences of the channels of u in PE[e] */
			set<unsigned int>::iterator pos;
			int j;
			for (set<TUndLink>::iterator l=e->first.begin(); l!=e->first.end(); l++)
				if ( (l->first->radio != NULL) && ((pos=uChannels.find(l->first->radio->channel)) != uChannels.end()) )
				{
					j=0;
					for (set<unsigned int>::iterator temp=uChannels.begin(); temp!=pos; temp++)
						j++;
					occurrences[j]++;
				}

			/* find the least used channel in PE[e] */
			j = min_element(occurrences.begin(), occurrences.end()) - occurrences.begin();
			pos = uChannels.begin();
			for (int i=0; i<j; i++) pos++;
			unsigned int k = *pos;
			assert(k>0);

			/* assign k to the first unassigned interface on v->info->id */
			v->info->radio[vChannels.size()].channel = k;

			/* set u->radio and v->radio */
			for (v->radio=&u->info->radio[0]; v->radio->channel != k; v->radio++);
			u->radio = &(v->info->radio[vChannels.size()]);
		}
		else
		{
			vector<int> occurrences(uChannels.size()+vChannels.size());
			for (vector<int>::iterator i=occurrences.begin(); i!=occurrences.end(); i++) *i=0;

			/* find the number of occurrences of the channels of u and v in PE[e] */
			set<unsigned int>::iterator pos;
			int j;
			for (set<TUndLink>::iterator l=e->first.begin(); l!=e->first.end(); l++)
				if (l->first->radio != NULL) 
				{
 					if ((pos=uChannels.find(l->first->radio->channel)) != uChannels.end())
					{
						j=0;
						for (set<unsigned int>::iterator temp=uChannels.begin(); temp!=pos; temp++)
							j++;
						occurrences[j]++;
					}
					else if ((pos=vChannels.find(l->first->radio->channel)) != vChannels.end())
					{
						j=uChannels.size();
						for (set<unsigned int>::iterator temp=vChannels.begin(); temp!=pos; temp++)
							j++;
						occurrences[j]++;
					}
				}

			/* find the least used channel in PE[e] */
			unsigned int k,k1;
			j = min_element(occurrences.begin(), occurrences.end()) - occurrences.begin();
			if ((unsigned int)j<uChannels.size())
			{
				pos = uChannels.begin();
				for (int i=0; i<j; i++) pos++;
				k = *pos;
				/* find the most used channel in vChannels */
				vector<int>::iterator temp=occurrences.begin();
				for (unsigned int i=0; i<uChannels.size(); i++) temp++;
				j = max_element(temp, occurrences.end()) - temp;
				pos = vChannels.begin();
				for (int i=0; i<j; i++) pos++;
				k1 = *pos;
				replace(T,v,u,k,k1);
			}
			else
			{
				pos = vChannels.begin();
				for (int i=uChannels.size(); i<j; i++) pos++;
				k = *pos;
				/* find the most used channel in uChannels */
				vector<int>::iterator temp=occurrences.begin();
				for (unsigned int i=0; i<uChannels.size(); i++) temp++;
				j = max_element(occurrences.begin(), temp) - occurrences.begin();
				pos = uChannels.begin();
				for (int i=0; i<j; i++) pos++;
				k1 = *pos;
				replace(T,u,v,k,k1);
			}
		}
	}

	return 0;
}
