#include "CAalgos.h"

#include <iostream>
#include <algorithm>
#include <cassert>

struct PEcard
{
	bool operator()(const set<TLinkItrs> &x, const set<TLinkItrs> &y)
	{
		return x.size() > y.size();
	}
};


/* Link l has been extracted and no common channels have been found
  Replaces channel k1 with k on a radio of the end node j */
void replace(Topology *T, TLinkItrs l, int j, unsigned int k, unsigned int k1)
{
	/* Find the radio of j which is assigned channel k1 */
// 	RadioIface *r;
// 	for (r=T->node[j].info->radio; r->channel!=k1; r++);
	/* replace k1 with k */
//	r->channel = k;
	for (vector<RadioIface>::iterator r=T->node[j].info->radio.begin(); r!=T->node[j].info->radio.end(); r++)
	  if (r->channel == k1)
	  {
	    r->channel = k;
	    break;
	  }


	/* assign k to the extracted link l */
	T->AssignChannel2Link(l, k );

	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			/* if u->v is a link leaving or entering node j and has been assigned channel k */
			if ( (u->info->id == j || v->info->id == j) && (v->radio!=NULL) && (v->radio->channel == k) )
			{
				set<unsigned int> uChannels, vChannels, CommonChannels;

// 				findChannels(u,uChannels);
// 				findChannels(v,vChannels);
				uChannels = u->info->findChannels();
				vChannels = v->info->findChannels();
				CommonChannels.clear();

				/* compute the intersection between uChannels and vChannels */
				set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(CommonChannels, CommonChannels.begin()));
	
				if (!CommonChannels.empty())
					T->AssignChannel2Link(TLinkItrs(u,v), *(CommonChannels.begin()) );
				else
				{
					int w = u->info->id == j ? v->info->id : u->info->id;
					replace(T,TLinkItrs(u,v),w,k,k1);
				}
			}
}



/* Given the initial graph, performs a channel assignment
   and returns the induced graph
*/
double LACA::Run (Topology *T)
{
//	Topology* T = new Topology(*Gi);     /* copy of the initial graph */

	/* for each link find the set of potentially interfering links */
	multimap<set<TLinkItrs>, TLinkItrs, PEcard> PE;		/* the map is sorted in descending order of |PE[e]| */

	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			{
				TLinkItrs e(u,v);
				set<TLinkItrs> PEe;
/* if i<->j potentially interfere with e, PEe only contains the pair <i,j> such that i->info->id > j->info->id */

				for (vector<TGraphNode>::iterator i=T->node.begin(); i!=T->node.end(); i++)
					for (list<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
						if (T->PotInterfLinks(TLinkItrs(i,j),TLinkItrs(u,v)))
							PEe.insert(TLinkItrs(i,j));

				PE.insert(pair<set<TLinkItrs>, TLinkItrs>(PEe, e));
			}

	for (multimap<set<TLinkItrs>, TLinkItrs, PEcard>::const_iterator e=PE.begin(); e!=PE.end(); ++e)
	{
		set<unsigned int> uChannels, vChannels, CommonChannels;
		/* end nodes of e */
		vector<TGraphNode>::iterator u=e->second.first;
		list<TGraphNeighbor>::iterator v=e->second.second;

// 		findChannels(u,uChannels);
// 		findChannels(v,vChannels);
		uChannels = u->info->findChannels();
		vChannels = v->info->findChannels();
		CommonChannels.clear();

		/* compute the intersection between uChannels and vChannels */
		set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(CommonChannels, CommonChannels.begin()));

		if (!CommonChannels.empty())
			T->AssignChannel2Link(e->second, *(CommonChannels.begin()) );
		else if ( (CommonChannels.empty()) && (uChannels.size()<u->info->numAvailRadios()) && (vChannels.size()<v->info->numAvailRadios()) )
		{
			vector<double> load(T->numAvailChannels+1);
			for (vector<double>::iterator i=load.begin(); i!=load.end(); i++) *i=0.;

			/* find the load of each channel in PE[e] */
			for (set<TLinkItrs>::iterator l=e->first.begin(); l!=e->first.end(); l++)
				if (l->second->radio != NULL)
					load[l->second->radio->channel]+=l->second->flow;

			/* find the least used channel in PE[e] */
			unsigned int k = min_element(++load.begin(), load.end()) - load.begin();
			assert(k>0);

			/* assign k to the link */
			T->AssignChannel2Link(e->second, k );
		}
		else if ( (CommonChannels.empty()) && (uChannels.size()<u->info->numAvailRadios()) && (vChannels.size()==v->info->numAvailRadios()) )
		{
			vector<double> load(vChannels.size());
			for (vector<double>::iterator i=load.begin(); i!=load.end(); i++) *i=0.;

			/* find the load of the channels of v in PE[e] */
			set<unsigned int>::iterator pos;
			int j;
			for (set<TLinkItrs>::iterator l=e->first.begin(); l!=e->first.end(); l++)
				if ( (l->second->radio != NULL) && ((pos=vChannels.find(l->second->radio->channel)) != vChannels.end()) )
				{
					j=0;
					for (set<unsigned int>::iterator temp=vChannels.begin(); temp!=pos; temp++)
						j++;
					load[j]+=l->second->flow;
				}

			/* find the least used channel in PE[e] */
			j = min_element(load.begin(), load.end()) - load.begin();
			pos = vChannels.begin();
			for (int i=0; i<j; i++) pos++;
			unsigned int k = *pos;
			assert(k>0);

			/* assign k to the link */
			T->AssignChannel2Link(e->second, k );
		}
		else if ( (CommonChannels.empty()) && (uChannels.size()==u->info->numAvailRadios()) && (vChannels.size()<v->info->numAvailRadios()) )
		{
			vector<double> load(uChannels.size());
			for (vector<double>::iterator i=load.begin(); i!=load.end(); i++) *i=0.;

			/* find the load of the channels of u in PE[e] */
			set<unsigned int>::iterator pos;
			int j;
			for (set<TLinkItrs>::iterator l=e->first.begin(); l!=e->first.end(); l++)
				if ( (l->second->radio != NULL) && ((pos=uChannels.find(l->second->radio->channel)) != uChannels.end()) )
				{
					j=0;
					for (set<unsigned int>::iterator temp=uChannels.begin(); temp!=pos; temp++)
						j++;
					load[j]+=l->second->flow;
				}

			/* find the least used channel in PE[e] */
			j = min_element(load.begin(), load.end()) - load.begin();
			pos = uChannels.begin();
			for (int i=0; i<j; i++) pos++;
			unsigned int k = *pos;
			assert(k>0);

			/* assign k to the link */
			T->AssignChannel2Link(e->second, k );
		}
		else
		{
			vector<double> load(uChannels.size()+vChannels.size());
			for (vector<double>::iterator i=load.begin(); i!=load.end(); i++) *i=0.;

			/* find the load of the channels of u and v in PE[e] */
			set<unsigned int>::iterator pos;
			int j;
			for (set<TLinkItrs>::iterator l=e->first.begin(); l!=e->first.end(); l++)
				if (l->second->radio != NULL) 
				{
 					if ((pos=uChannels.find(l->second->radio->channel)) != uChannels.end())
					{
						j=0;
						for (set<unsigned int>::iterator temp=uChannels.begin(); temp!=pos; temp++)
							j++;
						load[j]+=l->second->flow;
					}
					else if ((pos=vChannels.find(l->second->radio->channel)) != vChannels.end())
					{
						j=uChannels.size();
						for (set<unsigned int>::iterator temp=vChannels.begin(); temp!=pos; temp++)
							j++;
						load[j]+=l->second->flow;
					}
				}

			/* find the least used channel in PE[e] */
			unsigned int k,k1;
			j = min_element(load.begin(), load.end()) - load.begin();
			if ((unsigned int)j<uChannels.size())
			{
				pos = uChannels.begin();
				for (int i=0; i<j; i++) pos++;
				k = *pos;
				/* find the most used channel in vChannels */
				vector<double>::iterator temp=load.begin();
				for (unsigned int i=0; i<uChannels.size(); i++) temp++;
				j = max_element(temp, load.end()) - temp;
				pos = vChannels.begin();
				for (int i=0; i<j; i++) pos++;
				k1 = *pos;
				replace(T,e->second,v->info->id,k,k1);
			}
			else
			{
				pos = vChannels.begin();
				for (int i=uChannels.size(); i<j; i++) pos++;
				k = *pos;
				/* find the most used channel in uChannels */
				vector<double>::iterator temp=load.begin();
				for (unsigned int i=0; i<uChannels.size(); i++) temp++;
				j = max_element(load.begin(), temp) - load.begin();
				pos = uChannels.begin();
				for (int i=0; i<j; i++) pos++;
				k1 = *pos;
				replace(T,e->second,u->info->id,k,k1);
			}
		}
	}

	return 0;
}
