#ifndef WIMESH_PARSERS
#define WIMESH_PARSERS

#include <libxml++/libxml++.h>
#include "WMtypes.h"



class TopologyParser : public xmlpp::SaxParser
{
friend class Topology;

protected:
  // protected constructor: can only be invoked from within a Topology method
  TopologyParser (Topology* _T);
  virtual ~TopologyParser();

  unsigned int defaultNradios;
  float defaultDBgain;
  unsigned int defaultMinAngle;
  unsigned int defaultMaxAngle;

  // Node attributes may override default (Topology) attributes for that node's radios
  float nodeDBgain;
  unsigned int nodeMinAngle;
  unsigned int nodeMaxAngle;

  Topology* T;

  //overrides:
  virtual void on_start_document();
  virtual void on_end_document();
  virtual void on_start_element(const Glib::ustring& name,
                                const AttributeList& properties);
  virtual void on_end_element(const Glib::ustring& name);
  virtual void on_characters(const Glib::ustring& characters);
  virtual void on_comment(const Glib::ustring& text);
  virtual void on_warning(const Glib::ustring& text);
  virtual void on_error(const Glib::ustring& text);
  virtual void on_fatal_error(const Glib::ustring& text);
  
};



class WiMeshParser : public xmlpp::SaxParser
{
public:
  WiMeshParser ();
  virtual ~WiMeshParser();

  inline unsigned int getNumAvailChannels () { return numAvailChannels; }
  inline string getXmlTopoFile () { return xml_topo_file; }
  inline string getSerializedTopoFile () { return serialized_topo_file; }
  inline string getBuildGiMethod () { return buildGiMethod; }
  inline const list<pair<string,string> >& getBuildGiParams () { return buildGiParams; }
  inline string getPfrMethod () { return pfrMethod; }
  inline const list<pair<string,string> >& getPfrParams () { return pfrParams; }
  inline const map<int, NodePowerConsumptionProfile>& getNodePCprofile () { return nodePCprofile; }
  inline const map<int, RadioPowerConsumptionProfile>& getRadioPCprofile () { return radioPCprofile; }
  inline RndVar* getNodePCrv () { return nodePCrv; }
  inline double getNodePCseed () { return nodePCseed; }
  inline RndVar* getRadioPCrv () { return radioPCrv; }
  inline double getRadioPCseed () { return radioPCseed; }
  inline RndVar* getCDrv () { return CDrv; }
  inline double getCDseed () { return CDseed; }
  inline unsigned int getCDtype () { return CDtype; }
  inline string getCDid () { return CDid; }
  inline string getCAalgo() { return caAlgo; }
  inline const list<pair<string,string> >& getCAparams () { return caParams; }
  inline bool getpostCAopt () { return postCAopt; }
  inline string getSimulId () { return simulId; }


protected:
  unsigned int numAvailChannels;
  
  string xml_topo_file;
  string serialized_topo_file;

  string buildGiMethod;
  list<pair<string,string> > buildGiParams;

  string pfrMethod;
  list<pair<string,string> > pfrParams;

  map<int, NodePowerConsumptionProfile> nodePCprofile;
  map<int, RadioPowerConsumptionProfile> radioPCprofile;

  RndVar* nodePCrv;
  double nodePCseed;
  RndVar* radioPCrv;
  double radioPCseed;

  RndVar* CDrv;
  double CDseed;
  unsigned int CDtype;
  string CDid;

  string caAlgo;
  list<pair<string,string> > caParams;

  bool postCAopt;

  string simulId;

  //overrides:
  virtual void on_start_document();
  virtual void on_end_document();
  virtual void on_start_element(const Glib::ustring& name,
                                const AttributeList& properties);
  virtual void on_end_element(const Glib::ustring& name);
  virtual void on_characters(const Glib::ustring& characters);
  virtual void on_comment(const Glib::ustring& text);
  virtual void on_warning(const Glib::ustring& text);
  virtual void on_error(const Glib::ustring& text);
  virtual void on_fatal_error(const Glib::ustring& text);
  
};


#endif //WIMESH_PARSERS
