#ifndef WIMESH_PFR
#define WIMESH_PFR


#include "rndvars.h"
#include <cstring>
#include <list>


// Forward declarations
class Topology;    // WMtypes.h


class PFR_method
{
public:
  /* static method to create a new object of a derived class.
   * NOTE This method must be updated every time a new derived class is added
   */
  static PFR_method* Create (string name, const list<pair<string,string> >& params);
  
  virtual ~PFR_method() {}
  
  virtual void pfr_calc (Topology* T) = 0;

  const char* getStr() { return config_str; }

protected:
  void setStr(const char* str) { config_str = new char[strlen(str)+1]; strcpy(config_str,str); }
  
private:
  char* config_str;
};


class PFR_maxflow : public PFR_method
{
public:
  PFR_maxflow () { setStr("pfr_Maxflow"); }
  
  virtual void pfr_calc (Topology* T);
};


class PFR_allpairsMF : public PFR_method
{
public:
  PFR_allpairsMF () { setStr("pfr_AllpairsMF"); }
  
  virtual void pfr_calc (Topology* T);
};


class PFR_Raniwala : public PFR_method
{
public:
  PFR_Raniwala (RndVar* rv, double d) : Bwdr(rv), seed(d) { setStr(); }
  
  PFR_Raniwala (const list<pair<string,string> >& params);
  
  virtual void pfr_calc (Topology* T);
  
private:
  void setStr ();

  RndVar* Bwdr;
  double seed;
};


class PFR_Random : public PFR_method
{
public:
  PFR_Random (double _max, double _seed) : max(_max), seed(_seed) { setStr(); }
  
  PFR_Random (const list<pair<string,string> >& params);
  
  virtual void pfr_calc (Topology* T);
  
private:
  void setStr();

  double max;
  double seed;
};


/* The set of demands can be specified by means of either a random variable or a set of fixed values */
class PFR_MinMaxUtot : public PFR_method
{
public:
  PFR_MinMaxUtot (RndVar* rv, double _seed, bool _ksp, float _alpha, int _k, bool _capacity) 
  : DemVar(rv), seed(_seed), ksp(_ksp), alpha(_alpha), k(_k), capacity(_capacity) { setStr(); }
  
  PFR_MinMaxUtot (const list<pair<string,string> >& params);
  
  virtual void pfr_calc (Topology* T);
  
private:
  void setStr();

  RndVar* DemVar;
  double seed;
  bool ksp;
  float alpha;
  int k;
  bool capacity;
};


class PFR_mmcf : public PFR_method
{
public:
  PFR_mmcf () { setStr("pfr_MMCF"); }
  
  virtual void pfr_calc (Topology* T);
};

#if 0
class PFR_ksp : public PFR_method
{
public:
  PFR_ksp (bool _ksp, float _alpha, int _k) : ksp(_ksp), alpha(_alpha), K(_k) {}
  
  virtual void pfr_calc (Topology* T);
  
private:
  bool ksp;
  float alpha;
  int K;
};
#endif

// prototype
// PFR_method* readPFRmethod();

#endif
