#include "subgraph.h"

SubGraph::SubGraph(Topology* T) : Graph("emptygraph.data")
{
  clear();
  
  m_nVertexNum = T->node.size();
  
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      // construct the vertices
      BaseVertex* start_vertex_pt = get_vertex(u->info->id);
      BaseVertex* end_vertex_pt = get_vertex(v->info->id);
      
      // add the edge weight (1 for now)
      m_mpEdgeCodeWeight[get_edge_code(start_vertex_pt, end_vertex_pt)] = 1;
      
      // update the fan-in or fan-out variables
      // Fan-in
      get_vertex_set_pt(end_vertex_pt, m_mpFaninVertices)->insert(start_vertex_pt);

      // Fan-out
      get_vertex_set_pt(start_vertex_pt, m_mpFanoutVertices)->insert(end_vertex_pt);
  }
    
  m_nVertexNum = m_vtVertices.size();
  m_nEdgeNum = m_mpEdgeCodeWeight.size();
}


SubGraph::SubGraph(const SubGraph& graph) : Graph(graph)
{
  for (BaseVertexPt2SetMapIterator i=m_mpFaninVertices.begin(); i!=m_mpFaninVertices.end(); i++)
    i->second = new set<BaseVertex*>(*(i->second));
  
  for (BaseVertexPt2SetMapIterator i=m_mpFanoutVertices.begin(); i!=m_mpFanoutVertices.end(); i++)
    i->second = new set<BaseVertex*>(*(i->second));

  m_stRemovedEdge.clear();
  m_stRemovedVertexIds.clear();
}


SubGraph::~SubGraph()
{
  m_mpFanoutVertices.clear();
  m_mpFaninVertices.clear();
  m_vtVertices.clear();
  m_mpVertexIndex.clear();
}


void SubGraph::removeEdge(int u, int v) {
  m_mpEdgeCodeWeight.erase(get_edge_code(get_vertex(u),get_vertex(v)));
  get_vertex_set_pt(get_vertex(v), m_mpFaninVertices)->erase(get_vertex(u));
  get_vertex_set_pt(get_vertex(u), m_mpFanoutVertices)->erase(get_vertex(v));
  
  m_nEdgeNum = m_mpEdgeCodeWeight.size();
}


void SubGraph::removeVertex(int u)
{
  // remove incoming edges
  BaseVertexPt2SetMapIterator pos = m_mpFaninVertices.find(get_vertex(u));
  
  if (pos != m_mpFaninVertices.end()) {
    for (set<BaseVertex*>::iterator v=pos->second->begin(); v!=pos->second->end(); v++)
      m_mpEdgeCodeWeight.erase(get_edge_code(*v,get_vertex(u)));
    pos->second->clear();
  }
  
  // remove outgoing edges
  pos = m_mpFanoutVertices.find(get_vertex(u));
  
  if (pos != m_mpFanoutVertices.end()) {
    for (set<BaseVertex*>::iterator v=pos->second->begin(); v!=pos->second->end(); v++)
      m_mpEdgeCodeWeight.erase(get_edge_code(get_vertex(u),*v));
    pos->second->clear();
  }

  m_nEdgeNum = m_mpEdgeCodeWeight.size();
}
