#include <stdio.h>
#include <fstream>
#include <iostream>
#include <cassert>
#include <numeric>
#include <algorithm>
#include <ext/functional>
#include "WMtypes.h"
#include "WMutils.h"
#include "newran.h"
#include <sys/time.h>
#include <math.h>


bool operator<(const list<TGraphNeighbor>::iterator& x, const list<TGraphNeighbor>::iterator& y) {
  return ( (x->info->id < y->info->id) || ((x->info->id == y->info->id) && (&(*x) < &(*y))) );
}

bool operator<(const TGraphNeighbor& x, const TGraphNeighbor& y) {
  return x.info->id < y.info->id;
}


// bool operator<(const TLinkItrs& l1, const TLinkItrs& l2) { 
//   return ( (l1.first < l2.first) || ((l1.first==l2.first) && (l1.second < l2.second)) );
// }

#if 0
/* skipComment skips all blank and comment lines, if present, at the current file position
 * A comment line starts with #
 * Returns -1 in case of EOF
 */
int skipComment(FILE *f)
{
  char tmp[256];
  int c;

  do
  {
    /* skip initial whitespaces, tabs and carriage returns */
    fscanf(f,"%255[ \t\n\r]",tmp);

    c=fgetc(f);
    if (c=='#')
      fscanf(f,"%255[^\n\r]",tmp);   /* skip the whole line */

  } while ((c=='#') && (c!=EOF));

  if (c!=EOF)
    ungetc(c,f);

  return c==EOF ? -1 : 1;
}

int skipComment (istream& is)
{
  char c;

  do
  {
    is >> c;
    
    if (c=='#')
      is.ignore (500,'\n');
    else if (is.eof())
      return -1;
    else
      is.unget();
  } while (c=='#');
  
  return 1;
}
#endif

double getSeed()
{
/* This function provides random number generators with a random seed taken by current time */
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return (0.49 * sin((double)tv.tv_usec) + 0.50);    //  0.01 <= seed <= 0.99
}


void ReadParams::readBool (void* pvar)
{
  if (value == "true" || value == "1")
    *((bool*)pvar) = true;
  if (value == "false" || value == "0")
    *((bool*)pvar) = false;
}

void ReadParams::readRndVar (void* pvar)
{
  stringstream ss(value);
  *((RndVar**)pvar) = new RndVar (ss);
}


void genL2R_Markov(char *file, Topology *T)
{
/*	MotherOfAll urng(getSeed());         // declare uniform random number generator*/
	MotherOfAll urng(0.3);         // declare uniform random number generator
	Random::Set(urng);                   // set urng as generator to be used
	Uniform U;
	
	ofstream f;
	f.open(file);

	f << "numNodes=" << T->node.size() << ";" << endl << endl;
	
	f << "index=[" << endl;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			f << u->info->id * T->node.size() + v->info->id + 1 << " ";
	f << "];" << endl << endl;
	
	f << "flow=[" << endl;
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
			f << v->flow << " ";
	f << "];" << endl << endl;
	
	// choose one destination node among the gateways
	//SumRandom* D = new SumRandom(U*T->gateway_list.size());
	//int dest = T->gateway_list[(int)floor(D->Next())]->id;
	//f << "dest=" << dest+1 << ";" << endl << endl;
	
	f << "dest=[";
	for (vector<int>::iterator dest=T->gateway_list.begin(); dest!=T->gateway_list.end(); dest++)
		f << *dest + 1 << " ";
	f << "];" << endl << endl;

	linkPrune_Bwd prune(0.);

	f << "HC=[";
	for (vector<int>::iterator dest=T->gateway_list.begin(); dest!=T->gateway_list.end(); dest++)
	{
		/* computes distances using Dijkstra's */
		vector<TNodeProp> state = T->Dijkstra(*dest, prune);

		for (unsigned int i=0; i<T->node.size(); i++)
			f << (int)state[i].distance << " ";
		f << endl;
	}
	f << "];" << endl;

	f.close();
}


int findIFindex(set<unsigned int> channelset, unsigned int ch)
{
	int pos=0;

	for (set<unsigned int>::iterator i=channelset.begin(); (i!=channelset.end()) && (*i != ch); i++, pos++);

	return pos;
}

void genTcl(char *file, Topology *T)
{
	char filename[100];
	sprintf(filename,"%s%s",file,".tcl");

	ofstream f;
	f.open(filename);
	vector<TNodeProp> state;

	/* compute the number of interfaces */
	int numIfs = 0;
	for (vector<TGraphNode>::iterator i=T->node.begin(); i!=T->node.end(); i++)
		numIfs += i->info->numAvailRadios();

	/* compute the set of channels used */
	set<unsigned int> channels;
	T->getChannelSet(channels);

	/* write to file */
	f << "set val(nn)             " << T->node.size() << "                          ;# number of mobilenodes" << endl
	<< "set val(ni)             " << channels.size() << "                          ;# number of channels" << endl << endl
	<< "# Create god (number of interfaces * number of channels)" << endl
	<< "create-god [expr $val(nn)*$val(ni)]" << endl << endl
	<< "# Create wireless channels" << endl;

	for (set<unsigned int>::iterator ch=channels.begin(); ch!=channels.end(); ch++)
		f << "set chan_(" << *ch << ") [new $val(chan)]" << endl;

/*	<< "for {set i 0} {$i < $val(ni)} {incr i} {" << endl
	<< "	set chan_($i) [new $val(chan)]" << endl
	<< "}" << endl << endl*/

	f << endl << "$ns_ node-config -channel $chan_(" << *(channels.begin()) << ") \\" << endl
	<< "		 -ifNum $val(ni)" << endl << endl
	<< "# Setup agent and routing to new interfaces (common to all nodes)" << endl
	<< "set inerrProc_ " << (char)34 << (char)34 << endl
	<< "set outerrProc_ " << (char)34 << (char)34 << endl
	<< "set FECProc_ " << (char)34 << (char)34 << endl
	<< "set propInstance_ [new $val(prop)]" << endl
	<< "set port [Node set rtagent_port_]" << endl << endl;

	for (vector<TGraphNode>::iterator i=T->node.begin(); i!=T->node.end(); i++)
	{
		assert(!i->info->radio.empty());
		// find the set of channels (<>0) this node uses
		set<unsigned int> nodech = i->info->findChannels();
// 		findChannels(i,nodech);

		f << "# ---------------" << endl
		<< "# Create node " << i->info->id << endl
		<< "# ---------------" << endl
		<< "$ns_ change-numifs " << nodech.size() << endl;

		int j=0;
		for (set<unsigned int>::iterator c=nodech.begin(); c!=nodech.end(); c++, j++)
			f << "$ns_ add-channel " << j << " $chan_(" << *c << ")" << endl;

		f << endl << "set node_(" << i->info->id << ") [$ns_ node]" << endl
		<< "set agent [$node_(" << i->info->id << ") set ragent_]" << endl << endl;
	}

	f << endl << "# Nodes are static" << endl;
	for (vector<TGraphNode>::iterator i=T->node.begin(); i!=T->node.end(); i++)
		f << "$node_(" << i->info->id << ") random-motion 0" << endl;

	f << endl << "# Node positions" << endl;
	for (vector<TGraphNode>::iterator i=T->node.begin(); i!=T->node.end(); i++)
		f << "$node_(" << i->info->id << ") set X_ " << i->info->xpos << endl
		<< "$node_(" << i->info->id << ") set Y_ " << i->info->ypos << endl
		<< "$node_(" << i->info->id << ") set Z_ 0.0" << endl << endl;

	f << endl << "# Traffic sources and sinks" << endl
	<< "set trPairsNum " << T->aggregation_list.size() * T->gateway_list.size() << endl << endl;
	int srcid = 0, snkid = 0;
	for (vector<int>::iterator src=T->aggregation_list.begin(); src!=T->aggregation_list.end(); src++)
		for (vector<int>::iterator snk=T->gateway_list.begin(); snk!=T->gateway_list.end(); snk++)
			f << "set trSource(" << srcid++ << ") " << *src << endl
			<< "set trSink(" << snkid++ << ") " << *snk << endl;

	f.close();

	sprintf(filename,"%s%s",file,"-L2Rtable.tcl");
	f.open(filename);

//#ifdef L2ROUTING
	f << endl << endl << "# Manually makes a routing table for `l2routing`" << endl
	<< "# [$node_(i) set ragent_] add-neighbor <node id> <interface index> <flow>" << endl;
	for (vector<TGraphNode>::iterator i=T->node.begin(); i!=T->node.end(); i++)
	{
// 		double ftot = 0.;
// 		for (vector<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
// 			ftot += j->uLink->flow;

		// find the set of channels (<>0) this node uses
		set<unsigned int> nodech = i->info->findChannels();
// 		findChannels(i,nodech);

		for (list<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
			f << "[$node_(" << i->info->id << ") set ragent_] add-neighbor " << j->info->id << " "
			<< findIFindex(nodech,j->radio->channel) << " " << j->flow << endl;
//			(ftot ? j->uLink->flow / ftot : 1. / i->info->numAvailRadios()) << endl;
	}

	/* computes distances using Dijkstra's */
	set<int> destinations;
	for (vector<int>::iterator i=T->aggregation_list.begin(); i!=T->aggregation_list.end(); i++)
		destinations.insert(*i);
	for (vector<int>::iterator i=T->gateway_list.begin(); i!=T->gateway_list.end(); i++)
		destinations.insert(*i);

	linkPrune_Bwd prune(0.);
	for (set<int>::iterator dst=destinations.begin(); dst!=destinations.end(); dst++)
	{
		state = T->Dijkstra(*dst, prune);
		f << endl << "# Destination: node " << *dst << endl << "# ---------------------" << endl
		<< "#[$node_(i) set ragent_] add-dest-rtentry <destination node> <distance>" << endl;
		for (unsigned int i=0; i<T->node.size(); i++)
			if ((int)i != *dst)
			{
				f << "[$node_(" << i << ") set ragent_] add-dest-rtentry " << *dst << " " << (int)state[i].distance << endl;
			}
		f << "#[$node_(i) set ragent_] add-neighbor-rtentry <destination node> <node id> <type>" << endl;
		for (vector<TGraphNode>::iterator i=T->node.begin(); i!=T->node.end(); i++)
			if (i->info->id != *dst)
			{
				for (list<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
				{
					f << "[$node_(" << i->info->id << ") set ragent_] add-neighbor-rtentry " << *dst << " "
					<< j->info->id << " " << (int)state[j->info->id].distance - (int)state[i->info->id].distance << endl;
				}
			}
	}
//#endif

	f.close();
}


/* prints a .h file to include in topogen.c (Hyacinth) */
void genHyacinthTopo(Topology *T)
{
	ofstream f;
	f.open("stefano.h");

	for (vector<TGraphNode>::const_iterator u=T->node.begin(); u!=T->node.end(); u++)
		f << "node_locations[" << u->info->id << "].x = " << (int)u->info->xpos << ";" << endl
		<< "node_locations[" << u->info->id << "].y = " << (int)u->info->ypos << ";" << endl;

	f << endl << "cur_node = " << T->node.size() << ";"
	<< endl << "sample_size = " << T->node.size() << ";" << endl;

	f.close();

	f.open("traffic");

	for (vector<int>::iterator src=T->aggregation_list.begin(); src!=T->aggregation_list.end(); src++)
		for (vector<int>::iterator snk=T->gateway_list.begin(); snk!=T->gateway_list.end(); snk++)
			f << *src << " " << *snk << " 300" << endl;

	f.close();

	f.open("nodeinfo.hyacinth");

	f << "# Traffic sources and sinks" << endl
	<< "set trPairsNum " << T->aggregation_list.size() * T->gateway_list.size() << endl << endl;
	int srcid = 0, snkid = 0;
	for (vector<int>::iterator src=T->aggregation_list.begin(); src!=T->aggregation_list.end(); src++)
		for (vector<int>::iterator snk=T->gateway_list.begin(); snk!=T->gateway_list.end(); snk++)
			f << "set trSource(" << srcid++ << ") " << *src << endl
			<< "set trSink(" << snkid++ << ") " << *snk << endl;

	f.close();
}



void printCAStats(Topology *T, const char* name, bool verbose)
{
	set<unsigned int> channelSet;
	double Umax, Umin, Umean, Ustdev;
	vector<pair<TLinkItrs,double> >* m;

	cout << "*** " << name << " ***  " << endl;

	if (verbose)
	{
		T->getChannelSet(channelSet);
		for (set<unsigned int>::iterator i=channelSet.begin(); i!=channelSet.end(); i++)
		cout << *i << "  ";
		cout << endl << "Average node degree " << T->getDegree() << endl;
		
		T->printTopology();
	}

	// verbosity 2 => print the Utot of all the links
	T->checkUtot(2, &m);

	vector<double> u(m->size());

	transform(m->begin(), m->end(), u.begin(), __gnu_cxx::select2nd<pair<TLinkItrs,double> >());

	Umax = *max_element(u.begin(), u.end());
	Umin = *min_element(u.begin(), u.end());
	Umean = accumulate(u.begin(), u.end(), 0.) / (double)u.size();
	
	/* compute the standard deviation */
	transform(u.begin(), u.end(), u.begin(), bind1st(minus<double>(),Umean));
	transform(u.begin(), u.end(), u.begin(), u.begin(), multiplies<double>());
	
	Ustdev = sqrt(accumulate(u.begin(), u.end(), 0.) / (double)u.size());

	cout << "Maximum Utot = " << Umax << endl;
	cout << "Average Utot = " << Umean << endl;
	cout << endl;

	/* the maximum Utot only is sent to the standard error */
	cerr << name << "\t" << Umean - Ustdev << "\t" << Umin << '\t' << Umax << '\t' << Umean + Ustdev << '\t' << T->countRadiosOff() << endl;
	
	delete m;
}



