#! /usr/bin/octave -qf


# Usage:
# ./compare.m <mean|max> <udp|tcp> pattern...
#
# example: ./compare.m mean tcp 00? 01*


function PrintHelpAndExit ( )
  printf ("Usage: calcStat.m <what> <pattern>... [<X>] \n\n");
  printf ("what is a single char: (t)hroughput, (d)elay, (p)acket loss\n");
  printf ("The optional parameter X is used to compute the empirical CDF\n");
  exit
endfunction


###########
## Begin ##
###########

if nargin < 2
  PrintHelpAndExit()
endif

argc = 1;
arguments = argv();

switch arguments{argc}(1)
  case {'t'}
    colno = 4;
  case {'d'}
    colno = 2;
  case {'p'}
    colno = 6;
  otherwise
    PrintHelpAndExit ()
endswitch

argc++;

pattern = cell(0);

while argc <= nargin
  str = arguments{argc};

  if (argc == nargin && isvector(str2num(str)))
    X = str2num(str);
  else
    pattern(size(pattern,2)+1) = str;
  endif

  argc++;
endwhile

files = glob (pattern);

if numel (files) == 0
  printf ("%s", "No file found matching the specified criteria\n")
  exit
endif

namelen = sprintf("%%-%ds",max(cellfun("length",files))+5);

printf("\nAverage values for each simulation ");

switch arguments{1}(1)
  case {'t'}
    printf("(Mb/s)\n\n");
  case {'d'}
    printf("(ms)\n\n");
  otherwise
    printf("\n\n");
endswitch


data = [];
m = [];

## for each matching configuration

for i = 1:size(files,1)

  tmp = load("-ascii", files{i})(:,colno);

  if arguments{1}(1) == 't'
    tmp = tmp / 1000;
  endif

  eval(strcat("printf(""",namelen,"%f\\n"", files{i}, mean(tmp) )"))
  m = [m; mean(tmp)];
  data = [data; tmp];

endfor

printf("\nStatistics for the values above\n\n");
if length(m) > 1
  printf("%-12s%f\n","1st_quart",empirical_inv(0.25,m));
else
  printf("%-12s%f\n","1st_quart",m(1));
endif
printf("%-12s%f\n","Min",min(m));
printf("%-12s%f\n","Max",max(m));
if length(m) > 1
  printf("%-12s%f\n","3rd_quart",empirical_inv(0.75,m));
else
  printf("%-12s%f\n","3rd_quart",m(1));
endif
printf("%-12s%f\n","Mean",mean(m));


if !exist("X","var")

  X = min(data): (max(data)-min(data))/10 :max(data);

endif

printf("\nEmpirical CDF over 1s intervals (x,Pr{M<=x})\n\n");

printf("%f   %f\n",[X; empirical_cdf(X,data)])


printf("\nAverage values for each 1s interval over all the simulations\n\n");

res = mean(reshape(data,size(tmp,1),[]), 2)

printf("\nAverage of the above values: %f\n\n",mean(res))
