#include "wimesh-rtproto-ns3.h"
#include "ns3/mesh-helper.h"
#include "WMtypes.h"
#include <iostream>

#ifdef HAVE_LAYER2_5
#include "ns3/olsr-helper.h"
#include "query-olsrd.h"
#endif

// defined in ChannelAssign/core/utils.cpp
int findIFindex(set<unsigned int> channelset, unsigned int ch);

WiMeshRtProtoNs3::WiMeshRtProtoNs3 (string _stack, bool useWiMeshModels)
  : WiMeshNs3 (useWiMeshModels)
#ifdef HAVE_LAYER2_5
  , useOlsr (false)
#endif
{
  // ignore invalid values
#ifdef HAVE_LAYER2_5
  if (_stack == "ns3::LayerStack" || _stack == "ns3::FlameStack" || _stack == "ns3::Dot11sStack")
#else
  if (_stack == "ns3::FlameStack" || _stack == "ns3::Dot11sStack")
#endif
    stack = _stack;  
  else
  {
    std::cerr << "Ignoring invalid stack type (" << _stack << "), using default ns3::Dot11sStack" << std::endl;
    stack = "ns3::Dot11sStack";
  }
}

void WiMeshRtProtoNs3::CreateNs3MeshDevices ()
{
  set<unsigned int> chset;
  set<unsigned int>::iterator ch;
  Ptr<WifiNetDevice> wnd;

  MeshHelper mesh_helper = MeshHelper::Default ();

  mesh_helper.SetMacType ();

  SetRemoteStationManager (mesh_helper);

//  mesh_helper.SetRemoteStationManager ("ns3::WiMeshWifiManager",
//                                 "WiMeshNs3Ptr", WiMeshNs3PtrValue (WiMeshNs3Ptr (this)));
//  mesh_helper.SetRemoteStationManager("ns3::ConstantRateWifiManager", "DataMode",  StringValue ("OfdmRate54Mbps") );

  mesh_helper.SetStackInstaller (stack);

  // create mesh devices
  for (vector<TGraphNode>::iterator u = topology->node.begin (); u != topology->node.end (); u++)
  {
//     findChannels (u, chset);
    chset = u->info->findChannels();
    mesh_helper.SetNumberOfInterfaces (chset.size ());
    // create the required number of devices for the mesh node
    // the resulting meshPointDevice (a pseudo interface) is added to meshPointDevices
    meshPointDevices.Add (mesh_helper.Install (wifiPhy, NodeContainer (meshnodes.Get (u->info->id))));
    // assign channels to devices
    ch = chset.begin ();
    for (uint32_t i = 0; i < meshnodes.Get (u->info->id)->GetNDevices (); ++i)
    {
      wnd = meshnodes.Get (u->info->id)->GetDevice (i)->GetObject<WifiNetDevice> ();
      // if it is a WifiNetDevice
      if (wnd != 0){
	wnd->GetPhy ()->SetChannelNumber (*(ch++) * channelsep);
	meshdevices.Add(wnd);
      }
    }
  }
  
  //install IP stack and olsr routing protocol
  InternetStackHelper internetStack;
  #if 0
  if(stack=="ns3::LayerStack" && useOlsr){
  	OlsrHelper ols;
  	Ipv4StaticRoutingHelper staticRouting;
  	Ipv4ListRoutingHelper list;
  	list.Add (staticRouting, 0);
  	list.Add (ols, 10);
  	internetStack.SetRoutingHelper (list);
  }
  #endif
  
  internetStack.Install (meshnodes);
  Ipv4AddressHelper address;
  address.SetBase ("10.1.1.0", "255.255.255.0");
  Ipv4InterfaceContainer interfaces = address.Assign (meshPointDevices);
  
}

void WiMeshRtProtoNs3::ConfigureForwarding ()
{
#ifdef HAVE_LAYER2_5
  if (stack == "ns3::LayerStack")
  {
    ConfigureL2_5Links ();
    
    if (!useOlsr)
      ConfigureL2_5Topology ();
    else
      Simulator::Schedule(Seconds(30), &QueryOlsrd, meshnodes);
  }
#endif

  SetRoutes();
}

#ifdef HAVE_LAYER2_5
void WiMeshRtProtoNs3::EnableOlsr ()
{
  useOlsr = true;
}


void WiMeshRtProtoNs3::ConfigureL2_5Links ()
{
  // NOTE è uguale a ConfigureLinksWiMesh senza la parte sul RemoteManager
  set<unsigned int> nodech;
  set<unsigned int> neighborch;
  Ptr<layer::LayerProtocol> layer;

  // adding routing table entries
  for (vector<TGraphNode>::const_iterator i = topology->node.begin (); i != topology->node.end (); i++)
  {
    // find the set of channels (<>0) this node uses
//     findChannels(i,nodech);
    nodech = i->info->findChannels();

    // add links to nodes
    for (list<TGraphNeighbor>::const_iterator j = i->neighbor.begin (); j != i->neighbor.end (); j++)
    {
      //find the channels used by the neighbor
//       findChannels(j, neighborch);
      neighborch = j->info->findChannels();
      //main address of neighbor
      layer::LayerAddress neigh=Mac48Address::ConvertFrom(meshnodes.Get(j->info->id)->GetDevice(1)->GetAddress());

      //address of the neighbor's interface
      layer::LayerAddress neigh_ifn=Mac48Address::ConvertFrom(meshnodes.Get(j->info->id)->GetDevice(findIFindex(neighborch, j->radio->channel)+1)->GetAddress());
  
      std::cout<<"Adding neighbor " << j->info->id << " with main address " << neigh << " and interface address " << neigh_ifn << " to node " << i->info->id  << " on radio channel " << j->radio->channel << " with flow " << j->flow << " rate " << j->rate << " and power " << TxPowerLevel(j->txpower) << "\n";      

      layer = meshnodes.Get (i->info->id)->GetDevice (0)->GetObject<layer::LayerProtocol>();
      layer->AddNeighborLink (neigh, neigh_ifn, findIFindex(nodech,j->radio->channel), j->flow, j->rate, DbmToW(TxPowerLevel(j->txpower)) );
    }
  }
}

void WiMeshRtProtoNs3::ConfigureL2_5Topology ()
{
  // NOTE è uguale a ConfigureTopologyWiMesh
  Ptr<layer::LayerProtocol> layer;
  
  /* computes distances using Dijkstra's */
  set<int> destinations;
  for (vector<int>::iterator i=topology->aggregation_list.begin(); i!=topology->aggregation_list.end(); i++)
    destinations.insert(*i);
  for (vector<int>::iterator i=topology->gateway_list.begin(); i!=topology->gateway_list.end(); i++)
    destinations.insert(*i);
  
  linkPrune_Bwd prune(0.);
  
  // Add the destinations route entries
  for (set<int>::iterator dst=destinations.begin(); dst!=destinations.end(); dst++){

    //get address of destination
    layer::LayerAddress dest=Mac48Address::ConvertFrom(meshnodes.Get(*dst)->GetDevice(1)->GetAddress());
    vector<TNodeProp> state = topology->Dijkstra(*dst, prune);
    
    // add the destinations
    for (unsigned int i=0; i<topology->node.size(); i++)
      if ((int)i != *dst)
	{
	  layer = meshnodes.Get (i)->GetDevice(0)->GetObject<layer::LayerProtocol>();
	  layer->AddDestination(dest, (int)state[i].distance);
	}
    
    // Add the neighbor entries (NeighborRtEntry-s) for each destination
    // for each destination the type of each neighbor is given (tuple: destination neighbor_main_address type) 
    for (vector<TGraphNode>::iterator i=topology->node.begin(); i!=topology->node.end(); i++)
      if (i->info->id != *dst)
      {
	for (list<TGraphNeighbor>::iterator j=i->neighbor.begin(); j!=i->neighbor.end(); j++)
	{
	  // neighbor main address
	  layer::LayerAddress neighbor=Mac48Address::ConvertFrom(meshnodes.Get(j->info->id)->GetDevice(1)->GetAddress());
	  
	  int type=(int)state[j->info->id].distance - (int)state[i->info->id].distance;
	  
	  std::cout<<"Adding to node " << i->info->id << " for destination " << dest << " neighbor " << neighbor << " of type " << type<<".\n";
	  
	  layer = meshnodes.Get(i->info->id)->GetDevice(0)->GetObject<layer::LayerProtocol>();
	  layer->AddNeighborRtEntry(dest, neighbor, type);
	}
      }
  }
  
  
}
#endif

void WiMeshRtProtoNs3::SetRoutes()
{
// add the routing rules required to reach the nodes attached to the gateways
  Ipv4StaticRoutingHelper ipv4RoutingHelper;
  Ptr<Ipv4StaticRouting> staticRouting;
  
  for (TDemands::iterator d=demands.begin(); d != demands.end(); d++){

	// get the aggregator and gateway associated to the "demand" 
	int dst,src;
	src=d->first.first;
	dst=d->first.second;
	
	// get the ip address of the mesh interface of the gateway
	Ptr<Ipv4> ipv4 = meshnodes.Get (dst)->GetObject<Ipv4> ();
	Ipv4Address gateway_addr = ipv4->GetAddress (1, 0).GetLocal ();
	
	// get the static routing object associated to the aggregator
	Ptr<Node> agg = meshnodes.Get(src);
	Ptr<Ipv4> ipv4Aggr=agg->GetObject<Ipv4>();	
	staticRouting = ipv4RoutingHelper.GetStaticRouting (ipv4Aggr);

	// create the strings with the addresses of the point-to-point link
	stringstream host, prefix;
	prefix<< "192.168."<<dst;
	host << prefix.str() << ".1";
	
	// add the routing rule
	std::cout<<"Adding a routing rule to reach " << host.str() << " through " << gateway_addr  << " on node " << src <<  "." << std::endl;
	staticRouting->AddHostRouteTo (Ipv4Address (host.str().c_str ()), gateway_addr, 1);
  }
}


void WiMeshRtProtoNs3::DoReceiveFinalDataFailed(int me, Mac48Address neighbor)
{
#ifdef HAVE_LAYER2_5
  if (stack=="ns3::LayerStack")
  {
    Ptr<layer::LayerProtocol> layer = meshnodes.Get(me)->GetDevice(0)->GetObject<layer::LayerProtocol>();
    layer->ChangeNeighborStatus(neighbor, false);
  }
#endif
}



void WiMeshRtProtoNs3::ChangeTopology(Topology* T, int* changes)
{
  /*
   * switch channels and update IP addresses
   */
  vector<multiset<unsigned int> > newChSet, oldChSet;

  int res = topology->countRadioChanged (T, oldChSet, newChSet);
  
  if (changes)
    *changes = res;
  
  for (int u = 0; u < (int)T->node.size(); u++)
  {
    // do the channel switches
    multiset<unsigned int>::iterator newCh = newChSet[u].begin();
    for (multiset<unsigned int>::iterator oldCh = oldChSet[u].begin(); oldCh != oldChSet[u].end(); oldCh++, newCh++)
      DoChannelSwitch (u, *oldCh, *newCh);      
  }

  // the topology pointed to by WiMeshNs3 is the new one.
  // Keeping updated the topology member is required because, after a channel switch,
  // the WiMeshWifiManager tries to find the link rate. If the WiMesh topology is not updated,
  // a TGraphNeighbor with the new channel is not found and the basic rate (6Mbps) is used
  topology = T;
}

