#include <WMtypes.h>

#include "wimesh-wifi-phy.h"
#include "ns3/mobility-model.h"
#include "ns3/wifi-remote-station-manager.h"
#include "ns3/wifi-phy.h"

// defined in topology.cpp
float gain(float x);


using namespace ns3;

NS_OBJECT_ENSURE_REGISTERED (WiMeshPropagationLossModel);
NS_OBJECT_ENSURE_REGISTERED (WiMeshErrorRateModel);

TypeId 
WiMeshPropagationLossModel::GetTypeId (void)
{
  static TypeId tid = TypeId ("WiMeshPropagationLossModel")
    .SetParent<Object> ()
    .AddConstructor<WiMeshPropagationLossModel> ()
    ;
  return tid;
}


double 
WiMeshPropagationLossModel::DoCalcRxPower (double txPowerDbm,
                                          Ptr<MobilityModel> a,
                                          Ptr<MobilityModel> b) const
{

  double distance = a->GetDistanceFrom (b);
  
  //std::cout<<"POWER IN " <<txPowerDbm <<"\n";

  //gain from WiMesh topology.cpp 
  double g=gain(distance);  

  double pr = 10 * log10 (g);
  //NS_LOG_DEBUG ("distance="<<distance<<"m, attenuation coefficient="<<pr<<"dB");

  //std::cout<<"CALCULATING LOSS FOR DISTANCE " << distance << " " << g << endl;
  //std::cout<<"POWER OUT " <<txPowerDbm + pr <<"\n";

  return txPowerDbm + pr;
}


int64_t
WiMeshPropagationLossModel::DoAssignStreams (int64_t stream)
{
  return 0;
}

TypeId 
WiMeshErrorRateModel::GetTypeId (void)
{
  static TypeId tid = TypeId ("WiMeshErrorRateModel")
    .SetParent<ErrorRateModel> ()
    .AddConstructor<WiMeshErrorRateModel> ()
    ;
  return tid;
}

double 
WiMeshErrorRateModel::GetChunkSuccessRate (WifiMode mode, double snr, uint32_t nbits) const
{
  double rate=mode.GetDataRate()/(1024*1024); //get the rate in Mbit/s
  
  unsigned int i=0;
  bool found = false;
  double successRate=0.;
  
  static const double BOLTZMANN = 1.3803e-23;
  // Nt is the power of thermal noise in W
  double Nt = BOLTZMANN * 290.0 * mode.GetBandwidth ();
  
  while (i<NUMRATES && !found)
	if (availTXrates[i] >= rate)
		found = true;
	else
		i++;

//	assert (found);
	if (!found)
	  i = NUMRATES - 1;
  
  if (snr>=SINRthresholds[i]/Nt)
    successRate=1.0;
 
  //NS_LOG("data_date="<<rate<<"Mb,SINRthreshold["<<i<<"]="<<SINRthresholds[i]<<",SuccessRate="<<successRate);

  //std::cout<<"Phy rate " <<rate << " " << mode <<endl;
  //std::cout<<"SNR " << snr <<endl;
  //std::cout<<"Threshold " << i << " " << SINRthresholds[i]/Nt << endl;
  //std::cout<<"Success rate " << successRate << endl;
  //return 1.;
  return successRate;
}


