# find coin libs
#
# exports:
#
#   CoinLibs_FOUND
#   CoinLibs_INCLUDE_DIRS
#   CoinLibs_LIBRARIES
#
#   IpOpt_FOUND
#   IpOpt_INCLUDE_DIRS
#   IpOpt_LIBRARIES
#

include(FindPkgConfig)
include(FindPackageHandleStandardArgs)

set (CoinLibs    osi coinutils cbc cgl clp vol)
set (CoinLibsCap Osi CoinUtils Cbc Cgl Clp Vol)

# Use pkg-config to get hints about paths
pkg_check_modules(PC_COIN REQUIRED ${CoinLibs})


foreach (lib ${CoinLibs})

  string (TOUPPER ${lib} upper_lib)

  list (FIND CoinLibs ${lib} index)
  list (GET CoinLibsCap index lib_cap)

  # Include dir
  find_path(COIN_${upper_lib}_INCLUDE_DIR
    NAMES ${lib_cap}Config.h ${lib_cap}Volume.hpp 
    PATHS ${PC_COIN_${lib}_INCLUDEDIR}
  )

  # Finally the library itself
  find_library(COIN_${upper_lib}_LIBRARY
    NAMES ${lib_cap}
    PATHS ${PC_COIN_${lib}_LIBDIR}
  )

  FIND_PACKAGE_HANDLE_STANDARD_ARGS(Coin_${lib_cap} DEFAULT_MSG COIN_${upper_lib}_LIBRARY COIN_${upper_lib}_INCLUDE_DIR)


endforeach()

  if(PC_COIN_FOUND)
    set(CoinLibs_LIBRARIES ${PC_COIN_LIBRARIES} CACHE INTERNAL "CoinLibs libraries")
    set(CoinLibs_INCLUDE_DIRS ${PC_COIN_INCLUDE_DIRS} CACHE INTERNAL "CoinLibs include dirs")
    set(CoinLibs_FOUND yes)
  else()
    set(CoinLibs_LIBRARIES)
    set(CoinLibs_INCLUDE_DIRS)
    set(CoinLibs_FOUND no)
  endif()

# search for the optional IpOpt library

pkg_check_modules (PC_IPOPT ipopt)

# Include dir
find_path(IPOPT_INCLUDE_DIR
  NAMES IpoptConfig.h
  PATHS ${PC_IPOPT_INCLUDEDIR}
)

# Finally the library itself
find_library(IPOPT_LIBRARY
  NAMES ipopt
  PATHS ${PC_IPOPT_LIBDIR}
)

FIND_PACKAGE_HANDLE_STANDARD_ARGS(IpOpt DEFAULT_MSG IPOPT_LIBRARY IPOPT_INCLUDE_DIR)

  if(PC_IPOPT_FOUND)
    set(IpOpt_LIBRARIES ${PC_IPOPT_LIBRARIES} CACHE INTERNAL "IpOpt libraries")
    set(IpOpt_INCLUDE_DIRS ${PC_IPOPT_INCLUDE_DIRS} CACHE INTERNAL "IpOpt include dirs")
    set(IpOpt_FOUND yes)
  else()
    set(IpOpt_LIBRARIES)
    set(IpOpt_INCLUDE_DIRS)
    set(IpOpt_FOUND no)
  endif()
